"""
:Author: Juti Noppornpitak

This package contains the session controller used with the web controller and socket handler.
"""

from imagination.decorator.validator import restrict_type
from tori.session.repository.base    import Base

class Controller(object):
    """
    A session controller for the controller (request handler).
    """
    @restrict_type(Base)
    def __init__(self, session_repository, id):
        self._repository = session_repository
        self._id         = id

    @property
    def id(self):
        """ Administrated Session ID

        :return: str
        """
        return self._id

    def delete(self, key):
        """ Delete the data
        :param key: data key
        :type  key: str
        """
        self._repository.delete(self._id, key)

    def get(self, key):
        """ Retrieve the data

        :param key: data key
        :type  key: str
        :return: the data stored by the given key
        """
        return self._repository.get(self._id, key)

    def set(self, key, content):
        """ Define the data

        :param key: data key
        :type  key: str
        :param content: data content
        """
        self._repository.set(self._id, key, content)

    def reset(self):
        """ Clear out all data of the administrated session """
        self._repository.reset(self._id)



"""
Common Handler
==============

:Author: Juti Noppornpitak
:Status: Stable, Internal Only
:Last Update: |today|

This module contains code commonly used by request handlers and web socket handlers.
"""

from tori.centre             import services
from tori.session.generator  import GuidGenerator
from tori.session.controller import Controller

class Handler(object):
    _guid_generator = GuidGenerator()

    def __init__(self):
        """
        Handler Decorator Class used with Tornado's Handler-based classes

        .. note:: This should be moved to Controller and WebSocket.
        """
        self._session = None

    def component(self, name, fork_component=False):
        """
        Get the (re-usable) component from the initialized Imagination
        component locator service.

        :param `name`:           the name of the registered re-usable component.
        :param `fork_component`: the flag to fork the component
        :return:                 module, package registered or ``None``
        """

        if not services.has(name):
            return None

        return services.fork(name)\
            if   fork_component\
            else services.get(name)

    @property
    def session(self):
        if not self.component('session'):
            return None

        if self._session:
            return self._session

        cookie_key = 'ssid'

        ssid = self.get_secure_cookie(cookie_key)\
            if   self._can_use_secure_cookie()\
            else self.get_cookie(cookie_key)

        if not ssid:
            ssid = self._guid_generator.generate()

            if self._can_use_secure_cookie():
                self.set_secure_cookie(cookie_key, ssid, expires_days=1)
            else:
                self.set_cookie(cookie_key, ssid, expires_days=1)

        self._session = Controller(self.component('session'), ssid)

        return self._session

    def _can_use_secure_cookie(self):
        """
        Check if the secure cookie is enabled.

        :rtype: boolean

        .. note::
            This only works with any classes based from :class:`tornado.webRequestHandler`
            and :class:`tornado.websocket.WebSocketHandler`.

        """
        return 'cookie_secret' in self.settings\
            and self.settings['cookie_secret']


class Base(object):
    """
    Basic Session Entity is the base entity representing the data in thte session.
    """

    def __init__(self, session_id, key, content):
        self._session_id = session_id
        self.key         = key
        self.content     = content

    @property
    def session_id(self):
        return self._session_id


"""
:Author: Juti Noppornpitak
:Status: Testing
"""

from tori.session.entity.document import Document
from tori.session.repository.base import Base

class Collection(Base):
    """
    Session Controller using MongoDB

    :param `collection`: :class:`tori.db.odm.collection.Repository`

    .. note::

        This is based on :class:`tori.session.repository.base.Base`. All
        non-implementing methods in the parent class are implemented. This is
        only compatible with :class:`tori.session.entity.document.Document`.

    """

    def __init__(self, collection):
        Base.__init__(self)

        self.collection = collection

    def __get(self, id, key):
        return self.collection.filter_one(session_id=id, key=key)

    def delete(self, id, key):
        self.collection.delete_by_criteria(session_id=id, key=key)

    def get(self, id, key):
        data = self.__get(session_id=id, key=key)

        return data and data.content or None

    def registered(self, id):
        data = self.collection.filter_one(session_id=id)

        return len(data) > 0

    def reset(self, id):
        self.collection.delete_by_criteria(session_id=id)

    def set(self, id, key, content):
        data = self.__get(id, key)

        if not data:
            data = Document(id, key, content)

            self.collection.post(data)

            return

        data.content = content

        self.collection.put(data)


from tori.session.entity.base import Base
from tori.db.entity         import entity

@entity
class Document(Base):
    """
    Entity Session Entity is made to use with a Repository session repository (MongoDB).
    """
    def __init__(self, session_id, key, content, _id=None):
        Base.__init__(self, session_id, key, content)

        self._id = _id



"""
Remote Procedure Call Module
============================

:Author: Juti Noppornpitak
:Status: Stable/Testing
:Last Update: |today|
"""

import json
import time

from tori.data.serializer   import ArraySerializer
from tori.socket.websocket import WebSocket

class UncallableException(RuntimeError): pass

class Remote(object):
    """ RPC Request

    :param method:  the name of the method
    :type  method:  str
    :param id:      the request ID (default with unix timestamp)
    :param data:    method parameters
    :type  data:    dict
    :param service: the ID of the registered component/service (optional)
    :type  service: str
    """
    def __init__(self, method, id=None, data=None, service=None):
        self.id      = id or time.time()
        self.data    = data or None
        self.method  = method
        self.service = service or None

    def call(self):
        """ Execute the request

        :return: the result of the execution
        """
        remote_call = self.service.__getattribute__(self.method)

        if not callable(remote_call):
            raise UncallableException('The request method is not callable.')

        return remote_call(**self.data) if self.data else remote_call

class Response(object):
    """ RPC Response

    :param result: the result from RPC
    :param id: the response ID
    """
    def __init__(self, result, id):
        self.id     = id
        self.result = result

class ErrorResponse(object):
    def __init__(self, reason, description, feedback):
        self.reason      = reason
        self.description = description
        self.feedback    = feedback

class Interface(WebSocket):
    """ Remote Interface

    Extends from :class:`tori.socket.websocket.WebSocket`
    """

    def on_message(self, message):
        """
        :type message: str or unicode

        The parameter ``message`` is supposed to be in JSON format:

        .. code-block:: javascript

            {
                ["id":      unique_id,]
                ["service": service_name,]
                ["data":    parameter_object,]
                "method":  method_name
            }

        When the service is not specified, the interface will act as a service.

        """

        data     = json.loads(message)
        remote   = Remote(**data)
        response = None

        if remote.service:
            remote.service = self.component(remote.service)
        else:
            remote.service = self

        try:
            response = Response(remote.call(), remote.id)
        except AttributeError as e:
            response = ErrorResponse('The method does not exist.', e.message, data)

        simplified_response = ArraySerializer.instance().encode(response)
        simplified_response.update({
            'service': data['service'],
            'method':  data['method']
        })

        self.write_message(
            json.dumps(simplified_response)
        )



# -*- coding: utf-8 -*-

"""
:Author: Juti Noppornpitak

This package is used for rendering.
"""

from os import path, getcwd
import re

from jinja2         import Environment, FileSystemLoader, PackageLoader
from tori.centre    import settings as AppSettings
from tori.exception import *

from tori.template.repository import Repository

class Renderer(object):
    """
    The abstract renderer for Tori framework.

    .. warning::
        This is a non-working renderer. To use the built-in renderer (with
        Jinja2), try :class:`DefaultRenderer`. Otherwise, you should be
        expecting :class:`tori.exception.FutureFeatureException`.
    """
    def __init__(self, *args, **kwargs):
        raise FutureFeatureException("Need to implement.")

    def render(self, template_path, **contexts):
        """
        Render a template with context variables.

        :param template_path: a path to the template
        :type template_path: string or unicode
        :param contexts: a dictionary of context variables.
        :rtype: string or unicode

        Example::

            renderer = Renderer()
            renderer.render('dummy.html', appname='ikayaki', version=1.0)

        """
        raise FutureFeatureException("Need to implement.")

class DefaultRenderer(Renderer):
    """
    The default renderer with Jinja2

    :param `referers`: the template module path (e.g., com.shiroyuki.view)
                        or multiple base paths of Jinja templates based on the
                        current working directory.

    For example::

        # Instantiate with the module path.
        renderer = DefaultRenderer('app.views')

        # Instantiate with multiple base paths of Jinja templates.
        renderer = DefaultRenderer('/opt/app/ui/template', '/usr/local/tori/module/template')

    """

    def __init__(self, *referers):
        if len(referers) == 0:
            raise RendererSetupError('Require either one resource module or multiple file paths to the templates.')

        debug_mode = 'debug' in AppSettings and AppSettings['debug'] or False

        self.name     = ':'.join(referers)
        self.referers = referers
        self.loader   = self._get_package_loader() \
            if   (len(self.referers) == 1 and not path.exists(self.referers[0])) \
            else self._get_filesystem_loader()
        self.storage  = Environment(
            loader      = self.loader,
            trim_blocks = not debug_mode,
            auto_reload = debug_mode,
            extensions  = [
                'jinja2.ext.do',
                'jinja2.ext.i18n',
                'jinja2.ext.loopcontrols',
                'jinja2.ext.autoescape',
            ]
        )

    def _get_filesystem_loader(self):
        """
        Get the file-system loader for the renderer.

        :rtype: FileSystemLoader
        """
        for location in self.referers:
            if path.exists(location):
                continue

            backup_location = path.join(getcwd(), location)

            # unused logic
            #if backup_location:
            #    continue

            raise RendererSetupError('{} or {} is not found on this system.'.format(location, backup_location))

        return FileSystemLoader(self.referers)

    def _get_package_loader(self):
        """
        Get the package loader for the renderer.

        :rtype: PackageLoader
        """

        module_name_chunks       = re.split('\.', self.name)
        module_name              = '.'.join(module_name_chunks[:-1])
        template_sub_module_name = module_name_chunks[-1]

        if len(module_name_chunks) <= 1:
            raise RendererSetupError('Could not instantiate the package loader. (%s)' % '.'.join(module_name_chunks))

        return PackageLoader(module_name, template_sub_module_name)

    def render(self, template_path, **contexts):
        """
        See :meth:`Renderer.render` for more information.
        """

        template = self.storage.get_template(template_path)

        return template.render(**contexts)


from tori.session.repository.base   import Base   as BaseSessionAdapter
from tori.session.repository.file   import File   as FileSessionAdapter
from tori.session.repository.memory import Memory as InMemorySessionAdapter
from tori.session.repository.xredis import Redis  as RedisSessionAdapter


# -*- coding: utf-8 -*-

"""
:Author: Juti Noppornpitaks
:Restriction: Internal Use Only
"""

from tori.exception import *

class Repository(dict):
    """
    The template repository used by Rendering Service.

    :param class_reference: class reference
    :type class_reference:  tori.template.service.RenderingService

    .. note::
        This class is designed as a strict-type collection and may be refactored
        to the common area later on.
    """
    def __init__(self, class_reference):
        self._class_reference = class_reference

    def get(self, renderer_name):
        """
        Retrieve the renderer by name.

        :param renderer_name: the name of the renderer.
        :type renderer_name: string or unicode
        :rtype: tori.template.renderer.Renderer
        """
        if renderer_name not in self:
            raise RendererNotFoundError

        return self[renderer_name]

    def set(self, renderer):
        """
        Register the renderer.

        :return: self
        """
        if not isinstance(renderer, self._class_reference):
            raise UnsupportedRendererError('Expected an instance of %s but received %s.' % (self._class_reference, type(renderer)))

        self[renderer.name] = renderer

        return self



from sqlalchemy       import Column
from sqlalchemy.types import Integer, PickleType, String

from tori.db.wrapper           import Entity
from tori.session.entity.base import Base

class Database(Base, Entity):
    """
    DB Session Entity is made to use with a DB session repository.
    """

    __tablename__ = 'tori_session'

    id          = Column(Integer, primary_key=True)
    _session_id = Column(String(128), index=True)
    key         = Column(String(128), index=True)
    content     = Column(PickleType)


from tori.session.repository.base import Base

class Memory(Base):
    """ In-memory Session AbstractRepository """

    def __init__(self):
        self._storage = {}

    def delete(self, id, key):
        if not self.has(id, key):
            return

        del self._storage[id][key]

    def get(self, id, key, auto_close=True):
        if not self.has(id, key):
            return None

        return self._storage[id][key]

    def has(self, id, key):
        return id in self._storage and key in self._storage[id]

    def registered(self, id):
        return id in self._storage

    def reset(self, id):
        del self._storage[id]

    def set(self, id, key, content):
        if not self.registered(id):
            self._storage[id] = {}

        self._storage[id][key] = content


from tori.exception import *

class Base(object):
    """ The Base AbstractRepository """

    def __init__(self):
        pass

    def delete(self, id, key):
        """
        Delete a session key.

        :param `id`:  session ID
        :param `key`: session data key

        .. note:: This method is not implemented in :class:`tori.session.repository.base.Base`.
        """
        raise FutureFeatureException

    def generate(self):
        """
        Generate the session ID

        :return: Session ID as a string
        """
        if 'cookie_secret' not in AppSettings:
            raise SessionError('This session component could not not be initialized.')

        session_id = None

        while not session_id or self.registered(session_id):
            session_id = self.guidGenerator.generate()

        return session_id

    def get(self, id, key):
        """
        Set the given content to a session key.

        :param `id`:  session ID
        :param `key`: session data key
        :return:      session data content

        .. note:: This method is not implemented in :class:`tori.session.repository.base.Base`.
        """
        raise FutureFeatureException

    def has(self, id, key):
        """
        Set the given content to a session key.

        :param `id`:  session ID
        :param `key`: session data key
        :return:      ``True`` if there exists the data.
        """
        return self.get(session_id, key) is not None

    def registered(self, id):
        """
        Set the given content to a session key.

        :param `id`: session ID
        :return:     ``True`` if the session ID is registered.

        .. note:: This method is not implemented in :class:`tori.session.repository.base.Base`.
        """
        raise FutureFeatureException

    def reset(self, id):
        """
        Remove all data associated to the specified session.

        :param `id`:  session ID

        .. note:: This method is not implemented in :class:`tori.session.repository.base.Base`.
        """
        raise FutureFeatureException

    def set(self, id, key, content):
        """
        Set the given content to a session key.

        :param `id`:      session ID
        :param `key`:     session data key
        :param `content`: session data content
        :return:          session data content

        .. note:: This method is not implemented in :class:`tori.session.repository.base.Base`.
        """
        raise FutureFeatureException



# -*- coding: utf-8 -*-

"""
:Author: Juti Noppornpitak

..note::
    The documentation of this module yet conforms with the standard documenting
    style used by the official Python documenters.
"""

# Standard libraries
import json
import os
import sys

# Third-party libraries
from imagination.entity  import Entity as ImaginationEntity, CallbackProxy
from imagination.helper  import retrieve_module_path
from imagination.loader  import Loader as ImaginationLoader
from imagination.locator import Locator as ImaginationLocator
from imagination.helper.assembler import Assembler   as ImaginationAssembler
from imagination.helper.data      import Transformer as ImaginationTransformer
from kotoba             import load_from_file
from tornado.autoreload import watch
from tornado.ioloop     import IOLoop
from tornado.web        import Application as TornadoNormalApplication
import tornado.web
from tornado.wsgi       import WSGIApplication as TornadoWSGIApplication
from wsgiref            import handlers

# Internal libraries
from tori            import centre
from tori.centre     import settings as AppSettings
from tori.centre     import services as AppServices
from tori.common     import get_logger
from tori.data.base  import resolve_file_path
from tori.db.manager import ManagerFactory
from tori.exception  import *
from tori.navigation import *

class BaseApplication(object):
    """
    Interface to bootstrap a WSGI application with Tornado Web Server/Framework.
    This is the basic application class which practically does nothing. Please
    do not use this directly.

    :param `settings`: variable key-value parameters for extra settings to
    Tornado Web Server / Framework. Please read the web server documentation
    for more details.
    """

    _default_config_tree = {
        "session": {
            "class": "tori.session.repository.memory.Memory",
            "params": {}
        },
        "db": {
            "managers": {}
        }
    }

    def __init__(self, **settings):
        self._logger = get_logger('{}.{}'.format(__name__, self.__class__.__name__))

        self._hierarchy_level = len(self.__class__.__mro__) - 1

        self._ioloop = IOLoop.instance()

        settings.update(self._default_config_tree)

        # Setting for the application.
        self._settings          = settings
        self._settings['debug'] = False
        self._listening_addr    = '0.0.0.0'
        self._listening_port    = 8888

        if 'base_path' not in self._settings:
            # Get the reference to the calling function
            current_function    = sys._getframe(self._hierarchy_level)
            caller_function     = current_function.f_code
            reference_to_caller = caller_function.co_filename

            # Base path
            self._base_path = os.path.abspath(os.path.dirname(os.path.abspath(reference_to_caller)))
            self._base_path = 'static_path' in settings and settings['static_path'] or self._base_path

            self._settings['base_path'] = self._base_path

        self._base_path = self._settings['base_path']

        self._static_routing_setting = dict(path=self._base_path)
        self._routes = []

        # Enable the reverse reference
        disabled_core_reference = 'disabled_core_reference' in settings and settings['disabled_core_reference']

        if not disabled_core_reference and not centre.core:
            centre.core = self

    @property
    def base_path(self):
        return self._base_path

    def _update_routes(self, routes):
        """
        Update the routes.

        `routes` is the list of routes. See Tornado documentation on `tornado.web.Application` for more detail.
        """
        self._routes = routes

    def _activate(self):
        """
        Activate the backend application.
        """

        # Update the global settings.
        AppSettings.update(self._settings)

        # Instantiate the backend application.
        self._backend_app = TornadoNormalApplication(self._routes, **self._settings)

    def listen(self, port_number=None, bind_address=None):
        """
        Tell the app to listen on the given port number.

        `port_number` is an integer to indicate which port the application should be listen to.
        This setting is however only used in a standalone mode.
        """

        if port_number:
            self._listening_port = int(port_number)

        if bind_address:
            self._listening_addr = bind_address

        self._logger.debug('Listen at {} on port {}.'.format(self._listening_addr, self._listening_port))

        return self

    def add(self, callable_reference, timeout=500):
        self._ioloop.add_timeout(timeout, callable_reference)

    def start(self):
        """
        Start a server/service.
        """
        try:
            self._backend_app.listen(self._listening_port, self._listening_addr)

            self._logger.info("Start the application based at %s." % self._base_path)
            self._ioloop.start()
        except KeyboardInterrupt:
            self._logger.debug("Cleanly stopped.")

    def get_backbone(self):
        return self._backend_app

    def get_listening_port(self):
        return self._listening_port

class Application(BaseApplication):
    """
    Interface to bootstrap a WSGI application with Tornado Web Server.

    `settings` is a dictionary of extra settings to Tornado engine. For more
    information, please read Tornado documentation.
    """

    _registered_routing_types = ['controller', 'proxy', 'redirection', 'resource']
    _default_services         = [
        ('finder', 'tori.common.Finder', [], {}),
        ('renderer', 'tori.template.service.RenderingService', [], {}),
        ('session', 'tori.session.repository.memory.Memory', [], {}),
        ('routing_map', 'tori.navigation.RoutingMap', [], {}),
        ('db', 'tori.db.manager.ManagerFactory', [], {})
    ]

    _data_transformer         = ImaginationTransformer(ImaginationLocator())

    def __init__(self, configuration_location, **settings):
        BaseApplication.__init__(self, **settings)

        self._service_assembler = ImaginationAssembler(ImaginationTransformer(AppServices))
        self._config_main_path = os.path.join(self._base_path, configuration_location)
        self._config_base_path = os.path.dirname(self._config_main_path)

        self._config = load_from_file(self._config_main_path)

        # Initialize the routing map
        self._routing_map = RoutingMap()

        # Default properties
        self._scope = settings['scope'] if 'scope' in settings else None
        self._port  = 8000

        # Register the default services.
        self._register_default_services()

        # Add the main configuration to the watch list.
        watch(self._config_main_path)

        # Configure with the configuration files
        self._service_assembler.activate_passive_loading()

        for inclusion in self._config.children('include'):
            self._load_inclusion(inclusion)

        self._configure(self._config)

        self._prepare_db_connections()
        self._prepare_session_manager()

        self._service_assembler.deactivate_passive_loading()

        # Override the properties with the parameters.
        if 'port' in settings:
            self._port = settings['port']
            self._logger.info('Changed the listening port: %s' % self._port)

        # Update the routing map
        AppServices.get('routing_map').update(self._routing_map)

        # Normal procedure
        self._update_routes(self._routing_map.export())
        self.listen(self._port)
        self._activate()

    def _prepare_session_manager(self):
        config = self._settings['session']

        self._set_service_entity('session', config['class'], **config['params'])

    def _prepare_db_connections(self):
        db_config = self._settings['db']
        manager_config = db_config['managers']
        em_factory = AppServices.get('db')

        for alias in manager_config:
            url = manager_config[alias]['url']

            em_factory.set(alias, url)

            def callback(em_factory, db_alias):
                return em_factory.get(db_alias)

            callback_proxy = CallbackProxy(callback, em_factory, alias)

            AppServices.set('db.{}'.format(alias), callback_proxy)

    def _load_inclusion(self, inclusion):
        source_location = inclusion.attribute('src')

        if source_location[0] != '/':
            source_location = os.path.join(self._config_base_path, source_location)

        pre_config = load_from_file(source_location)

        self._configure(pre_config, source_location)

        watch(source_location)

        self._logger.info('Included the configuration from %s' % source_location)

    def _load_new_style_configuration(self, configuration):
        # Load the data directly from a JSON file.
        for inclusion in configuration.children('use'):
            source_location = inclusion.attribute('src')

            if source_location[0] != '/':
                source_location = os.path.join(self._config_base_path, source_location)

            with open(source_location) as f:
                decoded_config = json.load(f)

                self._override_sub_config_tree(self._settings, decoded_config)

                watch(source_location)

    def _override_sub_config_tree(self, original_subtree, modified_subtree):
        for key in modified_subtree:
            if key not in original_subtree:
                original_subtree[key] = modified_subtree[key]
                continue

            original_node_type = type(original_subtree[key])
            modified_node_type = type(modified_subtree[key])

            if original_node_type is dict:
                if modified_node_type != original_node_type:
                    raise ValueError('The overriding configuration tree does not align with the predefined one.')

                self._override_sub_config_tree(original_subtree[key], modified_subtree[key])

                continue

            original_subtree[key] = modified_subtree[key]

    def _configure(self, configuration, config_path=None):
        if len(configuration.children('server')) > 1:
            raise InvalidConfigurationError('Too many server configuration.')

        if len(configuration.children('routes')) > 1:
            raise InvalidConfigurationError('Too many routing configuration.')

        if len(configuration.children('settings')) > 1:
            raise InvalidConfigurationError('Too many setting groups (limited to 1).')

        self._load_new_style_configuration(configuration)

        # Then, load the legacy configuration. (Deprecated in 3.1)
        # Load the general settings
        for config in configuration.find('server config'):
            key  = config.attribute('key')
            kind = config.attribute('type')

            if not key:
                raise InvalidConfigurationError('Invalid server configuration key')

            self._settings[key] = self._data_transformer.cast(config, kind)

        # Set the cookie secret for secure cookies.
        client_secret = configuration.find('server secret')

        if client_secret:
            self._settings['cookie_secret'] = client_secret.data()

        # Set the port number.
        port = configuration.find('server port')

        if len(port) > 1:
            raise DuplicatedPortError()
        elif port:
            self._port = port.data()

        # Find the debugging flag
        if (configuration.find('server debug')):
            self._settings['debug'] = configuration.find('server debug').data().lower() == 'true'
            self._logger.info('Debug Mode: {}'.format('On' if self._settings['debug'] else 'Off'))

        for setting in configuration.children('settings').children('setting'):
            setting_name = setting.attr('name')

            if not setting_name:
                raise InvalidConfigurationError('A setting block does not specify the name. ({})'.format(config_path or 'primary configuration'))

            self._settings[setting_name] = setting.data()

        # Exclusive procedures
        self._register_imagination_services(configuration, config_path or self._config_base_path)
        self._map_routing_table(configuration)
        self._set_error_delegate(configuration)

    def _set_error_delegate(self, configuration):
        """ Set a new error delegate based on the given configuration file if specified. """
        delegate = configuration.find('server error').data()

        if delegate:
            self._logger.info('Custom Error Handler: %s' % delegate)
            tornado.web.ErrorHandler = ImaginationLoader(delegate).package

    def _register_default_services(self):
        for id, package_path, args, kwargs in self._default_services:
            self._set_service_entity(id, package_path, *args, **kwargs)

    def _register_imagination_services(self, configuration, base_path):
        """ Register services. """
        service_blocks = configuration.children('service')

        for service_block in service_blocks:
            service_config_path = service_block.data()

            # If the file path contains ':', treat the leading part for the full
            # module name and re-assembles the file path.

            config_file_path = resolve_file_path(service_config_path)

            if service_config_path[0] != '/':
                config_file_path = os.path.join(base_path, service_config_path)

            self._logger.info('Loading services from {}'.format(config_file_path))

            self._service_assembler.load(config_file_path)

    def _map_routing_table(self, configuration):
        """
        Update a routing table based on the configuration.
        """
        routing_sequences = configuration.children('routes')

        if not routing_sequences:
            return

        # Register the routes to controllers.
        for routing_sequence in routing_sequences:
            new_routing_map = RoutingMap.make(routing_sequence, self._base_path)

            self._routing_map.update(new_routing_map)

    def _make_service_entity(self, id, package_path, *args, **kwargs):
        """
        Make and return a service entity.

        *id* is the ID of a new service entity.

        *package_path* is the package path.

        *args* and *kwargs* are parameters used to instantiate the service.
        """
        loader = ImaginationLoader(package_path)
        entity = ImaginationEntity(id, loader, *args, **kwargs)

        return entity

    def _set_service_entity(self, id, package_path, *args, **kwargs):
        """
        Set the given service entity.

        *id* is the ID of a new service entity.

        *package_path* is the package path.

        *args* and *kwargs* are parameters used to instantiate the service.
        """
        AppServices.set(id, self._make_service_entity(id, package_path, *args, **kwargs))

    def get_route(self, routing_pattern):
        """ Get the route. """
        return self._routing_map.find_by_pattern(routing_pattern)

class WSGIApplication(Application):
    """
    Interface to bootstrap a WSGI application with Apache WSGI module.

    This class is simply a clone to :class:`Application` except that it is
    specially customized for deploying the application in the WSGI embedded mode.
    """

    def __init__(self, configuration_location, **settings):
        Application.__init__(self, configuration_location, **settings)

    def _activate(self):
        """
        Activate the backend application.
        """

        # Update the global settings.
        AppSettings.update(self._settings)

        # Instantiate the backend application.
        self._backend_app = TornadoWSGIApplication(self._routes, **self._settings)

    def start(self):
        """
        Start the process.

        .. warning:: This method is only necessary for a Google App Engine.
        """
        handlers.CGIHandler().run(self.get_backbone())



"""
:Author: Juti Noppornpitak
:Status: Testing/Unstable
"""

from tori.session.entity.database import Database as Entity
from tori.session.repository.base import Base

class Database(Base):
    """
    Session Controller using SQLAlchemy's ORM

    :param `url`: SQLAlchemy-compatible connection URL

    .. note::

        This is based on :class:`AbstractSession`. All non-implementing methods
        in the parent class are implemented. This is only compatible with
        :class:`DbEntity`.

    """

    def __init__(self, db):
        Base.__init__(self)

        self.db = db

    def base_filter(self, id):
        return self.db.\
            query(Entity).\
            filter(Entity.session_id==id)

    def commit(self):
        self.db.session.commit()

    def delete(self, id, key):
        data = self.base_filter(id).\
            filter(Entity.key==key).\
            first()

        if not data:
            return

        self.session.delete(data)
        self.commit()

    def get(self, id, key):
        data = self.base_filter(id).\
            filter(Entity.key==key).\
            first()

        return data and data.content or None

    def registered(self, id):
        data = self.base_filter(id).all()

        return len(data) > 0

    def reset(self, id):
        data_list = self.base_filter(id).all()

        for data in data_list:
            self.session.delete(data)

        self.commit()

    def set(self, id, key, content):
        data = self.base_filter(id).\
            filter(Entity.key==key).\
            first()

        if not data:
            data = Entity(id, key, content)

            self.session.add(data)
        else:
            data.content = content

        self.commit()


from tori.cli.exception import InterfaceException

class Command(object):
    """ Abstract class for all Tori-based commands """
    def define_arguments(self, argument_parser):
        """ Define the arguments """
        raise InterfaceException('Arguments are not defined.')

    def execute(self, args):
        """ Execute the command """
        raise InterfaceException('The interface does not know how to execute the command.')


"""
Generic Web Socket Module
=========================

:Author: Juti Noppornpitak
:Status: Stable
:Last Update: |today|
"""

from tornado.websocket import WebSocketHandler as BaseWebSocket

from tori.centre             import services
from tori.session.generator  import GuidGenerator
from tori.session.controller import Controller as SessionController

class WebSocket(BaseWebSocket):
    """ Web Socket Handler with extension to session controller """

    _guid_generator = GuidGenerator()
    _channel_table = {}

    def __init__(self, *args, **kwargs):
        BaseWebSocket.__init__(self, *args, **kwargs)

        self._session = None

        self._channel_table[self.__hash__()] = self

    def component(self, name, fork_component=False):
        """
        Get the (re-usable) component from the initialized Imagination
        component locator service.

        :param `name`:           the name of the registered re-usable component.
        :param `fork_component`: the flag to fork the component
        :return:                 module, package registered or ``None``
        """

        if not services.has(name):
            return None

        return services.fork(name)\
        if   fork_component\
        else services.get(name)

    @property
    def session(self):
        """ Session Controller

        :rtype: tori.session.controller.Controller
        """
        if not self.component('session'):
            return None

        if self._session:
            return self._session

        cookie_key = 'ssid'

        ssid = self.get_secure_cookie(cookie_key)\
            if   self._can_use_secure_cookie()\
            else self.get_cookie(cookie_key)

        if not ssid:
            ssid = self._guid_generator.generate()

            if self._can_use_secure_cookie():
                self.set_secure_cookie(cookie_key, ssid)
            else:
                self.set_cookie(cookie_key, ssid)

        self._session = SessionController(self.component('session'), ssid)

        return self._session

    def _can_use_secure_cookie(self):
        """ Check if the secure cookie is enabled.

        :rtype: boolean

        .. note::
            This only works with any classes based from :class:`tornado.webRequestHandler`
            and :class:`tornado.websocket.WebSocketHandler`.

        """
        return 'cookie_secret' in self.settings\
        and self.settings['cookie_secret']

    def broadcast(self, message, scopes=[]):
        current_session_id = self.session.id

        for object_hash in self._channel_table:
            channel = self._channel_table[object_hash]

            if current_session_id == channel.session.id:
                continue

            channel.write_message(message)

    def close(self):
        pass

    def on_close(self):
        del self._channel_table[self.__hash__()]

        self.close()


import json
import hashlib
import urllib
from tori.session.repository.base import Base

class File(Base):
    """ In-memory Session AbstractRepository """

    def __init__(self, location):
        self._storage  = {}
        self._location = location

        try:
            with open(self._location) as f:
                self._storage.update(json.load(f))
        except IOError:
            pass

    def delete(self, id, key):
        if not self.has(id, key):
            return

        del self._storage[id][key]

        self._update_backup()

    def get(self, id, key, auto_close=True):
        if not self.has(id, key):
            return None

        return self._storage[id][key]

    def has(self, id, key):
        return id in self._storage and key in self._storage[id]

    def registered(self, id):
        return id in self._storage

    def reset(self, id):
        del self._storage[id]

        self._update_backup()

    def set(self, id, key, content):
        if not self.registered(id):
            self._storage[id] = {}

        self._storage[id][key] = content

        self._update_backup()

    def _update_backup(self):
        with open(self._location, 'w') as f:
            json.dump(self._storage, f)



# -*- coding: utf-8 -*-

"""
:Author: Juti Noppornpitak

This package contains an abstract controller (based on
:class:`tornado.web.RequestHandler`) and built-in controllers.
"""

import logging

from base64 import b64decode
from os import path as p
from re import match, sub

from jinja2.exceptions import TemplateNotFound
from tornado.web import HTTPError, RequestHandler

from tori.centre             import services, core
from tori.common             import get_logger
from tori.data.base          import ResourceEntity, resolve_file_path
from tori.exception          import *
from tori.template.renderer  import DefaultRenderer
from tori.session.generator  import GuidGenerator
from tori.session.controller import Controller as SessionController

class Controller(RequestHandler):
    """
    The abstract controller for Tori framework which uses Jinja2 as a template
    engine instead of the default one that comes with Tornado.
    """

    _guid_generator     = GuidGenerator()
    _template_base_path = None
    _template_engine    = None


    def __init__(self, *args, **kwargs):
        RequestHandler.__init__(self, *args, **kwargs)

        self._session = None

    @property
    def is_xhr(self):
        return 'X-Requested-With' in self.request.headers and self.request.headers['X-Requested-With'] == 'XMLHttpRequest'

    def component(self, name, fork_component=False):
        """
        Get the (re-usable) component from the initialized Imagination
        component locator service.

        :param `name`:           the name of the registered re-usable component.
        :param `fork_component`: the flag to fork the component
        :return:                 module, package registered or ``None``
        """

        if not services.has(name):
            return None

        return services.fork(name)\
            if   fork_component\
            else services.get(name)

    @property
    def session(self):
        """ Session Controller

        :rtype: tori.session.controller.Controller
        """
        if not self.component('session'):
            return None

        if self._session:
            return self._session

        cookie_key = 'ssid'

        ssid = self.get_secure_cookie(cookie_key)\
            if   self._can_use_secure_cookie()\
            else self.get_cookie(cookie_key)

        if not ssid:
            ssid = self._guid_generator.generate()

            if self._can_use_secure_cookie():
                self.set_secure_cookie(cookie_key, ssid)
            else:
                self.set_cookie(cookie_key, ssid)

        self._session = SessionController(self.component('session'), ssid)

        return self._session

    def _can_use_secure_cookie(self):
        """
        Check if the secure cookie is enabled.

        :rtype: boolean

        .. note::
            This only works with any classes based from :class:`tornado.webRequestHandler`
            and :class:`tornado.websocket.WebSocketHandler`.

        """
        return 'cookie_secret' in self.settings and self.settings['cookie_secret']

    def render_template(self, template_name, **contexts):
        """ Render the template with the given contexts.

            See :meth:`tori.renderer.Renderer.render` for more information on the parameters.
        """

        contexts['app'] = {
            'settings': self.settings,
            'request': self.request,
            'session': self.session.get,
            'path':    self.component('routing_map').resolve
        }

        output = self.template_engine.render(template_name, **contexts)

        if not output:
            raise UnexpectedComputationError('Detected the rendering service malfunctioning.')

        return output

    def render(self, template_name, **contexts):
        """ Render the template with the given contexts and push the output buffer.

            See :meth:`tori.renderer.Renderer.render` for more information on the parameters.
        """
        self.write(self.render_template(template_name, **contexts))

    @property
    def template_engine(self):
        """ Template Engine

            :rtype: tori.template.renderer.Renderer
            :raise RenderingSourceMissingError: if the template base path and the reverse core reference are not defined.

            .. versionchanged:: 3.0

                The exception will not be raised if the reverse core reference is defined.

            .. note::

                The reverse core reference is the first instance of
                :class:`tori.application.Application` in the process.
        """

        # If the rendering source isn't set, break the code.
        if not core and not self._template_base_path:
            raise RenderingSourceMissingError('The source of template is not identified. This method is disabled.')
        elif not self._template_base_path:
            self._template_base_path = p.join(core.base_path, 'templates')

        # If the rendering engine is not specified, use the default one.
        if not self._template_engine:
            self._template_engine = DefaultRenderer

        try:
            return self.component('renderer').use(self._template_base_path)
        except RendererNotFoundError:
            try:
                # When the renderer is not found. It is possible that the renderer is not yet
                # instantiated. This block of the code will do the lazy loading.
                renderer = self._template_engine(self._template_base_path)

                self.component('renderer').register(renderer)
            except ImportError as exception:
                raise RuntimeError(
                    'Unable to register the renderer for {} as {}'.format(
                        self._template_base_path,
                        exception.message
                    )
                )

            return renderer

class SimpleController(Controller):
    """ Simplified Request Controller """
    def rendering_prefix(self):
        return None

    def get(self, *args):
        path = args[0] if args else 'index'

        if '..' in path:
            raise HTTPError(403)

        if '/' == path[-1]:
            path = 'index'

        rendering_path = '{}.html'.format(path)

        if self.rendering_prefix():
            rendering_path = p.join(self.rendering_prefix(), rendering_path)

        try:
            rendered_data = self.render_template(rendering_path)
        except TemplateNotFound as e:
            raise HTTPError(404)

        if not rendered_data:
            raise HTTPError(404)

        self.write(rendered_data)

class RestController(Controller):
    """ Abstract REST-capable controller based on a single primary key. """
    def list(self):
        """ Retrieve the list of all entities. """
        self.set_status(405)

    def retrieve(self, key):
        """ Retrieve an entity with `key`. """
        self.set_status(405)

    def create(self):
        """ Create an entity. """
        self.set_status(405)

    def remove(self, key):
        """ Remove an entity with `key`. """
        self.set_status(405)

    def update(self, key):
        """ Update an entity with `key`. """
        self.set_status(405)

    def get(self, key=None):
        """ Handle GET requests. """
        if not key:
            self.list()
            return

        self.retrieve(key)

    def post(self, key=None):
        """ Handle POST requests. """
        if id:
            self.set_status(405)
            return

        self.create()

    def put(self, key=None):
        """ Handle PUT requests. """
        if not key:
            self.set_status(400)
            return

        self.update(key)

    def delete(self, key=None):
        """ Handle DELETE requests. """
        if not key:
            self.set_status(400)
            return

        self.remove(key)

class ErrorController(Controller):
    """Generates an error response with status_code for all requests."""
    def initialize(self, status_code):
        self.set_status(status_code)

    def prepare(self):
        raise HTTPError(self._status_code)

class ResourceService(Controller):
    """ Resource service is to serve a static resource via HTTP/S protocol. """

    _logger = get_logger('%s.ResourceService' % (__name__), logging.ERROR)

    _favicon_data = b64decode(''.join([
        'AAABAAEAEBAQAAAAAAAoAQAAFgAAACgAAAAQAAAAIAAAAAEABAAAAAAAgAAAAAAAAAAA',
        'AAAAEAAAAAAAAAAAAAAA/4QAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA',
        'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEAAAAAAAAAARAAAAAAAAABEQAAAAAAAAEREA',
        'AAAAAAARERAAAAAAABEREQAAAAAAEREREAAAAAARERERAAAAABEREREQAAAAEREREREA',
        'AAARERERERAAABEREREREQAAEREREREREAARERERERERABEREREREREQERERERERERF/',
        '/wAAP/8AAB//AAAP/wAAB/8AAAP/AAAB/wAAAP8AAAB/AAAAPwAAAB8AAAAPAAAABwAA',
        'AAMAAAABAAAAAAAA'
    ]))

    _patterns      = {}
    _pattern_order = []
    _cache_objects = {}

    _plugins            = {}
    _plugins_tag_name   = 'resource-service-plugin'
    _plugins_registered = False

    @staticmethod
    def add_pattern(pattern, base_path, enable_cache=False):
        """
        Add the routing pattern for the resource path prefix.

        :param pattern: a routing pattern. It can be a Python-compatible regular expression.

        :param base_path: a path prefix of the resource corresponding to the routing pattern.

        :param enable_cache: a flag to indicate whether any loaded resources need to be cached on the first request.
        """
        ResourceService._logger.debug('add URL pattern "%s" for "%s"' % (pattern, base_path))
        ResourceService._patterns[pattern] = {
            'base_path': resolve_file_path(base_path),
            'cacheable': enable_cache
        }
        ResourceService._pattern_order.append(pattern)

    def get(self, path=None):
        """
        Get a particular resource.

        :param path: blocks of path used to composite an actual path.

        .. note::
            This method requires refactoring.
        """
        resource = self._retrieve_resource_entity()

        if isinstance(resource, str):
            self.set_header('Content-Type', 'image/vnd.microsoft.icon')
            return self.finish(resource)
        elif resource.exists and resource.cacheable:
            self._cache_objects[self.request.uri] = resource
        elif not resource.exists:
            # Return HTTP 404 if the content is not found.
            self._logger.error('%s could not be found.' % resource.path)

            raise HTTPError(404)

        # Get the content type.
        self.set_header("Content-Type", resource.kind or 'text/plain')

        # Retrieve the plugins if registered.
        if not ResourceService._plugins and not ResourceService._plugins_registered:
            ResourceService._plugins = services.find_by_tag(
                ResourceService._plugins_tag_name
            )

        # Apply the plugin.
        for plugin in ResourceService._plugins:
            if plugin.expect(resource):
                resource = plugin.execute(resource)
            # End the iteration

        if resource.is_originally_dir:
            return self.redirect('{}/{}'.format(path, ResourceEntity.DEFAULT_INDEX_FILE))

        # Return the content.
        try:
            self.finish(resource.content)
        except Exception as e:
            raise HTTPError(500)

    def _retrieve_resource_entity(self):
        request_uri = self.request.uri
        path        = sub('\?.*$', '', request_uri)

        if request_uri in self._cache_objects:
            return self._cache_objects[request_uri]

        # If the request URI is already pre-calculated or fixed, load the
        # entity from the corresponding path.
        elif request_uri in ResourceService._patterns:
            pattern = ResourceService._patterns[request_uri]

            return self._create_resource_entity(pattern['base_path'], pattern['cacheable'])

        # When the resource is not loaded, try to get from the wildcard pattern.
        self._logger.debug('Retrieving from the wildcard pattern.')

        return self._get_resource_on_non_precalculated_pattern(path)

    def _create_resource_entity(self, real_path, cachable):
        return ResourceEntity(real_path, cachable)

    def _get_resource_on_non_precalculated_pattern(self, request_uri):
        for pattern in ResourceService._pattern_order:
            pattern_info = ResourceService._patterns[pattern]

            base_path = pattern_info['base_path']
            cachable  = pattern_info['cacheable']

            self._logger.debug('Comparing Pattern: %s' % pattern)

            matches = match(pattern, request_uri)

            if not matches:
                continue

            real_path = p.abspath(p.join(
                base_path,
                matches.groups()[0]
            ))

            self._logger.info('Real path: %s' % real_path)

            return self._create_resource_entity(real_path, cachable)

        if self.request.uri == '/favicon.ico':
            return self._favicon_data

        raise HTTPError(404)


from time import time

from tori.decorator.common import singleton
from tori.centre import settings as AppSettings
from tori.common import Enigma

@singleton
class GuidGenerator(object):
    def generate(self):
        key = '%s/%s' % (AppSettings['cookie_secret'], time())\
            if 'cookie_secret' in AppSettings\
            else str(time())

        return Enigma.instance().hash(key)



class InterfaceException(Exception):
    """ Interface Exception. Require implementation """

class TerminationSignal(RuntimeError):
    """ Termination Signal """

class CommandNotFound(RuntimeError):
    """ Command Not Found """

class NotConfigured(Exception):
    """ Not-configured exception """


# -*- coding: utf-8 -*-

"""
:Author: Juti Noppornpitak
:Restriction: Internal Use Only

This package contains the rendering service. This is a module automatically loaded by :class:`tori.application.Application`.
"""

from tori.exception           import *
from tori.template.renderer   import Renderer
from tori.template.repository import Repository

class RenderingService(object):
    """ The rendering service allows the access to all template repositories.

    This acts as a controller.

    :param renderer_class: a class reference of a renderer
    :type renderer_class: tori.template.renderer.Renderer
    :param repository_class: a class reference of a template repository
    :type repository_class: tori.template.repository.Repository
    """

    def __init__(self, renderer_class=Renderer, repository_class=Repository):
        self._repository = repository_class(renderer_class)

    def register(self, renderer):
        """
        Register a renderer.

        :param renderer: the renderer
        :type renderer: tori.template.renderer.Renderer
        :return: ``self``.
        """
        self._repository.set(renderer)

        return self

    def use(self, repository_name):
        """ Retrieve the renderer by name

        :param repository_name: the name of the repository
        :type  repository_name: str
        :rtype: tori.template.renderer.Renderer
        """
        return self._repository.get(repository_name)

    def render(self, repository_name, template_path, **contexts):
        """ Render a template from a repository *repository_name*.

        As this method acts as a shortcut and wrapper to the actual renderer for the given repository,
        see :meth:`tori.template.renderer.Renderer.render` for more information.

        :rtype: string
        """
        return self.use(repository_name).render(template_path, **contexts)



# -*- coding: utf-8 -*-

"""
:Author: Juti Noppornpitaks
:Purpose: Internal Use Only

The navigation module is designed specifically for the dependency-injectable Application.

Please note that the term *DOMElement* used on this page denotes any of :class:`yotsuba.kotoba.Kotoba`,
:class:`yotsuba.kotoba.DOMElements` and :class:`yotsuba.kotoba.DOMElement`.

Additionally, the parameter *route* for any methods mentioned on this page is an instance of *DOMElement*.
"""

# Standard libraries
from os import path
from re import compile as RegExp

# Third-party libraries
from imagination.loader import Loader
from kotoba.kotoba      import Kotoba
from tornado.web        import RedirectHandler

# Internal libraries
from tori.common    import get_logger
from tori.exception import *

class RoutingMap(object):
    """ Routing Map """

    def __init__(self):
        self._logger         = get_logger('{}.{}'.format(__name__, self.__class__.__name__))
        self._sequence       = []
        self._raw_map        = {}
        self._id_map         = {}
        self._final_sequence = None

    def register(self, route, force_action=False):
        """ Register a *route*. """

        if not isinstance(route, Route):
            raise InvalidInput("Expected a route.")

        # Register the pattern to prevent the duplicate routing.
        if not force_action and route.pattern in self._sequence:
            raise DuplicatedRouteError('The route pattern is already registered.')

        if route.pattern not in self._sequence:
            self._sequence.append(route.pattern)

        # Register the route to the map.
        self._raw_map[route.pattern] = route

        # Reset the final sequence.
        self._final_sequence = None

        # Register the route to the short ID map.
        if not force_action and route.id and route.id in self._id_map:
            raise DuplicatedRouteError('The route ID "{}" is already registered.'.format(route.id))

        self._id_map[route.id] = route

        self._logger.debug('Added route for "{}"'.format(route.pattern))

    def resolve(self, id, **params):
        """ Resolve the path by ID

            :param id str: the path ID
            :param params: the variables used in the routing pattern
        """

        if id not in self._id_map:
            raise KeyError('The route ID "{}" is not registered.'.format(id))

        route = self._id_map[id]

        if route.use_regexp:
            raise RuntimeError('The route is defined with a regular expression. It cannot be resolved.')

        return route.resolvable_pattern.format(**params)

    def find_by_pattern(self, routing_pattern):
        """ Get the route by *routing_pattern* where it is a string. """

        try:
            return self._raw_map[routing_pattern]
        except KeyError:
            raise RoutingPatternNotFoundError(
                'Failed to find the route for "{lookup}" in {availables}'.format(
                    lookup = routing_pattern,
                    availables = ', '.join(self._raw_map.keys())
                )
            )

    def export(self):
        """ Export the route map as a list of tuple representatives.

        :rtype: list
        """
        if not self._final_sequence:
            has_favicon = False

            self._final_sequence = []

            for routing_pattern in self._sequence:
                if routing_pattern == '/favicon.ico':
                    has_favicon = True

                self._final_sequence.append(self.find_by_pattern(routing_pattern).to_tuple())

            if not has_favicon:
                self._final_sequence.append(('/favicon.ico', StaticRoute._default_service))

        return self._final_sequence

    def update(self, other):
        for route in other._raw_map.values():
            self.register(route, True)

    @staticmethod
    def make(configuration, base_path=None):
        """
        Make a routing table based on the given *configuration*.

        :param `base_path`: is an optional used by :method `Route.make`:.
        """
        if not configuration:
            return

        if not isinstance(configuration, Kotoba):
            raise InvalidInput

        routing_map = RoutingMap()

        # Register the routes to controllers.
        for route_data in configuration.children():
            routing_map.register(Route.make(route_data, base_path))

        return routing_map

class Route(object):
    """The abstract class representing a routing directive.

    :param route: an instance of :class:`kotoba.kotoba.Kotoba` representing the route.
    """
    _re_wildcard_recur        = RegExp('(\*{2,})')
    _re_wildcard_non_recur    = RegExp('(\*{1})')
    _re_named_parameter       = RegExp('\{(?P<name>[^\}]+)\}')
    _registered_routing_types = ['controller', 'proxy', 'redirection', 'resource']

    def __init__(self, route_data):
        self._source  = route_data
        self._class   = None
        self._pattern = Route.get_pattern(self._source)
        self._resolvable_pattern = None
        self._id      = self._source.attribute('id')

        self._use_regexp = self._source.attribute('regexp') or False

        if self._use_regexp:
            self._use_regexp = self._use_regexp.lower() == 'true'

        # If the given pattern is not a regular expression, rewrite the routing pattern.
        if not self.use_regexp:
            self._resolvable_pattern = self._pattern
            self._pattern = self._re_wildcard_recur.sub('(.+)', self._pattern)
            self._pattern = self._re_wildcard_non_recur.sub('([^/]+)', self._pattern)
            self._pattern = self._re_named_parameter.sub('(?P<\g<name>>.+)', self._pattern)

    @property
    def id(self):
        return self._id

    @property
    def use_regexp(self):
        return self._use_regexp

    @property
    def pattern(self):
        return self._pattern

    @property
    def resolvable_pattern(self):
        return self._resolvable_pattern

    def type(self):
        """ Get the routing type.

        :rtype: str
        """
        return self.source().name()

    def source(self):
        """ Get the original data for the route.

        :rtype: str
        """
        return self._source

    def bean_class(self):
        """ Get the class reference for the route.

        :rtype: type
        """
        if not self._class and self.source().attribute('class'):
            self._class = Loader(self.source().attribute('class')).package

        return self._class

    @staticmethod
    def make(route_data, base_path=None):
        if not isinstance(route_data, Kotoba):
            raise InvalidInput('')

        route = None
        kind  = Route.get_type(route_data)

        # Create a route by type.
        if kind == 'controller':
            route = DynamicRoute(route_data)
        elif kind == 'resource' and base_path:
            route = StaticRoute(route_data, base_path)
        elif kind == 'resource' and not base_path:
            raise InvalidInput('base_path is not provided.')
        elif kind == 'redirection':
            route = RelayRoute(route_data)
        elif kind == 'proxy':
            raise FutureFeatureException('Routing a proxy is yet supported at the moment.')

        if not route:
            raise UnknownRoutingTypeError(kind)

        return route

    @staticmethod
    def get_type(route_data):
        """ Get the routing type for a given *route*. """
        if route_data.name() not in Route._registered_routing_types:
            raise RoutingTypeNotFoundError('')

        return route_data.name()

    @staticmethod
    def get_pattern(route_data):
        """ Get the routing pattern for a given *route*. """
        if not route_data.attribute('pattern'):
            raise RoutingPatternNotFoundError('The pattern is not given.')

        return route_data.attribute('pattern')

class DynamicRoute(Route):
    """ Dynamic route based on class Route handled by a controller. """

    def __init__(self, route):
        super(self.__class__, self).__init__(route)

        self._controller = None

    def controller(self):
        """ Get the controller. """
        if not self._controller:
            self._controller = self.bean_class()

        return self._controller

    def to_tuple(self):
        """ Convert the route to tuple. """
        return (self.pattern, self.controller())

class StaticRoute(Route):
    """
    Static routing directive based on :class:`Route` handled by a resource controller

    :param base_path: is a string indicating the base path for the static resource.
    """
    _base_path       = None
    _default_service = None

    def __init__(self, route, base_path):
        super(self.__class__, self).__init__(route)

        self._location = None

        if not self._base_path:
            self._base_path = base_path

        # Map the routing pattern to the actual location.
        self.service().add_pattern(self.pattern, self.location(), self.cache_enabled())

    def location(self):
        """ Get the location of the static resource/content. """

        if not self.source().attribute('location'):
            raise InvalidControllerDirectiveError('The location of the resource is missing.')

        if not self._location:
            self._location = self.source().attribute('location')

            secondary_location = path.join(self._base_path, self._location)

            # If the path of the resource is not an absolute path, treat the path as a relative path.
            if self._location[0] != '/' and path.exists(secondary_location):
                self._location = secondary_location

        return self._location

    def cache_enabled(self):
        """ Check whether the caching option is enabled. """

        _cache_enabled = self.source().attribute('cache')

        return _cache_enabled and _cache_enabled.lower() == 'true' or False

    def service(self):
        """ Get the resource service. """
        service = self.bean_class() or StaticRoute.default_service()

        return service

    def to_tuple(self):
        """ Convert the route to tuple. """
        return (self.pattern, self.service())

    @staticmethod
    def default_service():
        if not StaticRoute._default_service:
            StaticRoute._default_service = Loader('tori.controller.ResourceService').package

        return StaticRoute._default_service

class RelayRoute(Route):
    """ Relay routing directive based on :class:`Route` used for redirection """

    def __init__(self, route):
        super(self.__class__, self).__init__(route)

        self._destination = None

    def is_permanent(self):
        """ Check whether the relay route is permanent. """
        _is_permanent = self.source().attribute('permanent')
        return _is_permanent and _is_permanent.lower() == 'true' or False

    def destination(self):
        """ Get the relaying destination. """

        if not self._destination:
            self._destination = self.source().attribute('destination')

        if not self._destination:
            raise InvalidControllerDirectiveError('The destination is missing.')

        return self._destination

    def to_tuple(self):
        """ Convert the route to tuple. """

        return (
            self.pattern,
            RedirectHandler, {
                'url':       self.destination(),
                'permanent': self.is_permanent()
            }
        )



# -*- coding: utf-8 -*-
import argparse
import re
import sys
from imagination.decorator.validator import restrict_type
from tori.application   import Application
from tori.centre        import services
from tori.cli.exception import NotConfigured, TerminationSignal, CommandNotFound

class Console(object):
    """ Main Console

        The commands must be defined first with tag "command" and then any thing
        with prefix "command:". For example,

        .. code-block:: xml

            <!-- From https://github.com/nepteam/nep -->
            <entity
                id="command.db"
                class="neptune.command.Database"
                tags="command command:db">
                <param name="db" type="entity">db</param>
                <interception before="me" do="execute" with="init"/>
                <interception after="me" do="execute" with="clean_up"/>
            </entity>

        Then, the command will be referenced with anything after ":". From the
        previous example, the command "command:db" will be referred as "db".
    """
    def __init__(self, namespace=None):
        self.namespace = namespace
        self.app       = None

    @restrict_type(Application)
    def load(self, application):
        self.app = application

    def output(self, *args):
        output = ' '.join(args)
        print(output)

    def show_console_usage(self):
        if 'command' not in services._tag_to_entity_ids:
            self.output('{} has no sub commands registered.'.format(self.namespace))
            return

        self.output('USAGE: {} <command>'.format(self.namespace or sys.argv[0]))
        self.output('\nAvailable commands:')

        # This is a hack to get the list of IDs from Imagination Framework.
        command_ids = [entity_id for entity_id in services._tag_to_entity_ids['command']]
        command_desc_map = {}
        longest_cmd_length = 0

        for command_id in command_ids:
            # This is also a hack to get the entity (metadata) of the commands.
            entity = services.get_wrapper(command_id)
            for tag in entity.tags:
                if not re.search('^command:', tag):
                    continue

                alias  = re.sub('^[^:]+:', '', tag)

                if longest_cmd_length < len(alias):
                    longest_cmd_length = len(alias)

                command_doc = entity.loader.package.__doc__ or 'Unknown command'
                command_desc_map[alias] = command_doc.strip()

        format_string = '  {:<%d}{}' % (longest_cmd_length + 4)

        for id in command_desc_map:
            self.output(format_string.format(id, command_desc_map[id]))

    def handle(self):
        if not self.app:
            raise NotConfigured('Please load the configuration first.')

        if len(sys.argv) == 1:
            self.show_console_usage()
            raise TerminationSignal('List command.')

        command_id = sys.argv[1]

        command  = None
        commands = services.find_by_tag('command:{}'.format(command_id))

        if commands:
            command = commands[0]

        if not command:
            self.show_console_usage()
            raise CommandNotFound(command_id)

        # Reconfigure the parser.
        parser         = argparse.ArgumentParser(description='Console')
        subparsers     = parser.add_subparsers()
        command_parser = subparsers.add_parser(command_id, description=command.__class__.__doc__.strip())

        command.define_arguments(command_parser)
        command.execute(parser.parse_args())


try:
    from setuptools import setup
except:
    from distutils.core import setup

setup(
    name         = 'tori',
    version      = '3.0.3',
    description  = 'A collection of libraries, a micro web framework based on Tornado framework, and the ORM for MongoDB',
    license      = 'MIT',
    author       = 'Juti Noppornpitak',
    author_email = 'juti_n@yahoo.co.jp',
    url          = 'http://shiroyuki.com/projects/tori.html',
    packages     = [
        'tori',
        'tori.cli',
        'tori.data',
        'tori.db',
        'tori.db.driver',
        'tori.db.metadata',
        'tori.decorator',
        'tori.session',
        'tori.session.entity',
        'tori.session.repository',
        'tori.socket',
        'tori.template'
    ],
    classifiers   = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries'
    ],
    scripts          = ['bin/nest'],
    install_requires = ['imagination', 'kotoba', 'tornado', 'jinja2', 'pymongo']
)



from tori.decorator.common import singleton

@singleton
class ArraySerializer(object):
    MODE_STRICT   = 'strict'
    """ Re-raise all errors. """
    MODE_FORGIVEN = 'forgiven'
    """ Bypass all errors. """

    def __init__(self, max_depth=2, mode=MODE_FORGIVEN):
        self._max_depth = max_depth
        self._primitive_types = []
        self._string_types = []
        self._mode = mode

    def set_max_depth(self, max_depth):
        self._max_depth = max_depth

    def encode(self, data, stack_depth=0):
        if not isinstance(data, object):
            raise TypeError('The provided data must be an object')

        returnee = {}

        for name in dir(data):
            if name[0] == '_':
                continue

            value = data.__getattribute__(name)

            if callable(value):
                continue

            if value and not self._is_primitive_type(value):
                try:
                    if self._max_depth and stack_depth >= self._max_depth:
                        value = value.encode('utf8', 'ignore')
                    else:
                        value = self.encode(value, stack_depth + 1)
                except AttributeError as e:
                    if self._mode == ArraySerializer.MODE_STRICT:
                        raise RuntimeError(
                            'Error while processing property {name} of type {kind} ({original})'.format(
                                name = name,
                                kind = type(data),
                                original = e.message
                            )
                        )

                    continue

            returnee[name] = value

        return returnee

    def _is_primitive_type(self, value):
        if not self._primitive_types:
            self._primitive_types = self.default_primitive_types()

        return type(value) in self._primitive_types

    def _is_string(self, value):
        if not self._string_types:
            self._string_types = self.default_string_types()

        return type(value) in self._string_types

    def default_string_types(self, value):
        try:
            return [str, unicode]
        except NameError as exception:
            return [str]

    def default_primitive_types(self):
        try:
            return [int, float, str, list, dict, tuple, set, bool, unicode, long]
        except NameError as exception:
            return [int, float, str, list, dict, tuple, set, bool]


"""
:Author: Juti Noppornpitak

This package contains decorators for enhancing controllers.
"""
try:
    # Python 2.6+
    import httplib
except ImportError as error:
    # Python 3.3+
    import http.client as httplib

import traceback

from tori.exception import *

def _assign_renderer(class_reference, base_path, engine=None):
    """
    Assign the renderer to the given controller-type class.

    `class_reference` is a reference to a class type, not an instance of a class.

    The first argument called `source` is a name of the package containing templates and static contents.

    The second argument called `renderer_type` is a name of a renderer-type class used to override the default renderer.

    To use this, suppose we have a class called `DummyController` and later instantiate
    a variable `dummy_controller` as an instance of class `DummyController`. `class_reference`
    will be `DummyController`, not `dummy_controller`.

    Note that this method is not for direct use. You can invoke via `@renderer` or the
    configuration file only via `tori.application.Application`.

    Additionally, the arguments might be different if the renderer is not the default one.
    """
    class_reference._template_base_path = base_path
    class_reference._template_engine    = engine

    return class_reference

def renderer(*args, **kwargs):
    """ Set up the renderer for a controller.

    See :class:`tori.template.renderer.Renderer` for more information.
    """
    def inner_decorator(class_reference):
        if len(args) == 0:
            raise InvalidInput

        return _assign_renderer(class_reference, *args, **kwargs)

    return inner_decorator

def _enable_custom_error(controller, template_name, **contexts):
    def write_error(self, status_code, **kwargs):
        debug_info = []

        if self.settings.get("debug") and "exc_info" in kwargs:
            # in debug mode, try to send a traceback
            for line in traceback.format_exception(*kwargs["exc_info"]):
                debug_info.append(line)

        self.render(
            template_name,
            code       = status_code,
            message    = httplib.responses[status_code],
            debug_info = ''.join(debug_info),
            **contexts
        )

    controller.write_error = write_error

def custom_error(template_name, **contexts):
    """ Set up the controller to handle exceptions with a custom error page.

    .. note:: This decorator is to override the method ``write_error``.

    :param template_name: the name of the template to render.
    :type  template_name: string
    :param contexts:      map of context variables
    :type  contexts:      dict
    """
    def updated_controller(controller):
        _enable_custom_error(controller, template_name, **contexts)

        return controller

    return updated_controller


try:
    import cPickle as pickle
except:
    import pickle

import redis
from time import time

from tori.exception               import *
from tori.session.repository.base import Base

class Redis(Base):
    def __init__(self, prefix='tori/session', redis_client=None, use_localhost_as_fallback=True):
        Base.__init__(self)

        self._redis = redis_client
        self._prefix = prefix

        self._use_localhost_as_fallback = use_localhost_as_fallback

    @property
    def __api(self):
        if not self._redis and not self._use_localhost_as_fallback:
            raise InvalidInput('The Redis API (PIP: redis) must be provided.')
        elif not self._redis and self._use_localhost_as_fallback:
            pool        = redis.ConnectionPool(host='localhost', port=6379, db=0)
            self._redis = redis.Redis(connection_pool=pool)

        return self._redis

    def __compose_key(self, id, key):
        return '/'.join([self._prefix, id, key])

    def delete(self, id, key):
        actual_key = self.__compose_key(id, key)

        self.__api.delete(actual_key)

    def get(self, id, key):
        actual_key = self.__compose_key(id, key)
        content    = self.__api.get(actual_key)

        if content:
            content = pickle.loads(content)

        return content

    def registered(self, id):
        actual_key = self.__compose_key(id, '*')

        return self.__api.keys(actual_key)

    def reset(self, id):
        self.delete(id, '*')

    def set(self, id, key, content):
        actual_key = self.__compose_key(id, key)

        self.__api.set(actual_key, pickle.dumps(content))